from typing import Optional

from pydantic import ConfigDict

from .base import DirectusModel


class Role(DirectusModel):
    """
    Directus role model.
    """

    id: Optional[str] = None
    name: Optional[str] = None

    model_config = ConfigDict(collection="directus_roles")


class User(DirectusModel):
    """
    Directus user model.
    """

    id: Optional[str] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    avatar: Optional[str] = None
    description: Optional[str] = None
    email: Optional[str] = None
    role: Optional[str] | Optional[Role] = None
    status: Optional[str] = None
    title: Optional[str] = None
    token: Optional[str] = None

    model_config = ConfigDict(collection="directus_users")



import secrets
from typing import Union, Optional, List


RANDOM_STRING_CHARS = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"


def get_random_string(length=None, allowed_chars=RANDOM_STRING_CHARS):
    """
    Return a securely generated random string.

    The bit length of the returned value can be calculated with the formula:
        log_2(len(allowed_chars)^length)

    For example, with default `allowed_chars` (26+26+10), this gives:
      * length: 12, bit length =~ 71 bits
      * length: 22, bit length =~ 131 bits

    Origin: https://github.com/django/django/blob/3.2.23/django/utils/crypto.py
    """
    if length is None:
        length = 12
    return "".join(secrets.choice(allowed_chars) for i in range(length))


def parse_translations(all_translations: List[dict]) -> Union[dict[str, dict[str, str]], None]:
    if all_translations is None or not all_translations:
        return None

    return {
        translations['key']: {
            translation['languages_code']: translation['translation']
            for translation in translations['translations']
        } for translations in all_translations
    }



from __future__ import annotations

import asyncio
import json as jsonlib
from typing import (
    TYPE_CHECKING,
    Union, Optional, Type, List, Tuple,
    overload
)

import json_fix
import websockets
from pydantic import BaseModel

from py_directus.aggregator import Agg
from py_directus.directus_response import DirectusResponse
from py_directus.filter import F

# from py_directus.operators import AggregationOperators

if TYPE_CHECKING:
    from websockets import Data, WebSocketClientProtocol
    from py_directus import Directus


class DirectusRequest:
    """
    Class to manage request to the Directus API.
    """
    _lock = asyncio.Lock()
    _lock_2 = asyncio.Lock()

    def __init__(self, directus: 'Directus', collection: str,
                 collection_class: Optional[Union[Type[BaseModel], str]] = None):
        json_fix.fix_it()

        self.directus: 'Directus' = directus
        self.collection: str = collection
        self.params: dict = {}
        self.collection_class: Optional[Union[Type[BaseModel], str]] = collection_class

    @property
    def uri(self):
        if "directus_" in self.collection:
            return f"{self.directus.url}/{self.collection.replace('directus_', '')}"
        return f"{self.directus.url}/items/{self.collection}"

    def fields(self, *fields):
        self.params['fields'] = ",".join(fields)
        return self

    def filter(self, *args, **filters):
        """
        :param operator: The operator to use for the filter (Operators.Equals)
        :param logical_operator: The logical operator to use for the filter (LogicalOperators.And)
        :param filters: Multiple filters to use

        :return: The DirectusRequest object

        :example:
                .filter(Operators.Equals, LogicalOperators.Or, first_name="Panos", location=None) \
                .filter(Operators.Equals, last_name="Stavrianos") \
        """
        # Argument handling
        clean_args = list(filter(lambda x: isinstance(x, F), args))

        new_args_filter = None
        if clean_args:
            new_args_filter = clean_args[0]

            for fltr in clean_args[1:]:
                new_args_filter &= fltr

            if "filter" in self.params:
                self.params['filter'] = self.params['filter'] & new_args_filter
            else:
                self.params['filter'] = new_args_filter

        # Keyword argument handling
        # filter_param = Filter(operator, logical_operator, **filters)
        filter_param = F(**filters)

        if "filter" in self.params:
            self.params['filter'] = self.params['filter'] & filter_param
        else:
            self.params['filter'] = filter_param

        return self

    def sort(self, field, asc=True):
        if 'sort' not in self.params:
            self.params['sort'] = []
        self.params['sort'].append(f'{"" if asc else "-"}{field}')
        return self

    def search(self, search: Optional[Union[str, int]] = None):
        self.params['search'] = search
        return self

    def page(self, page: int = 1):
        self.params['page'] = page
        return self

    def limit(self, limit: int = -1):
        self.params['limit'] = limit
        return self

    def offset(self, offset: int = 0):
        self.params['offset'] = offset
        return self

    def aggregate(self, *args, **aggregates):
        """
        :param operator: The operator to use for the aggregation (AggregationOperators.Count)
        :param field: Field name uppon which the aggregation will be performed

        :return: The DirectusRequest object
        """

        # Argument handling
        clean_args = list(filter(lambda x: isinstance(x, Agg), args))

        new_args_aggregator = None
        if clean_args:
            new_args_aggregator = clean_args[0]

            for aggr in clean_args[1:]:
                new_args_aggregator += aggr

            if "aggregate" in self.params:
                self.params['aggregate'] = self.params['aggregate'] + new_args_aggregator
            else:
                self.params['aggregate'] = new_args_aggregator

        # Keyword argument handling
        aggregator_param = Agg(**aggregates)

        if "aggregate" in self.params:
            self.params['aggregate'] = self.params['aggregate'] + aggregator_param
        else:
            self.params['aggregate'] = aggregator_param

        return self

    def group_by(self, *fields):
        self.params['groupBy'] = ",".join(fields)
        return self

    def include_count(self):
        # NOTE: DEPRECATED, USE AGGREGATION INSTEAD
        self.params['meta'] = "*"
        return self

    async def read(
            self, id: Optional[Union[int, str]] = None, method: str = "search", cache: bool = False,
            as_task: bool = False
    ) -> DirectusResponse:
        """
        Request data.

        :param id: The id of the item to retrieve
        :param method: The method to use for the request (search, get)
        :param cache: Whether to use the cache or not
        :param as_task: Whether to add the request to the tasks list or not (for batch requests)

        :return: The DirectusResponse object

        IMPORTANT: cache and as_task cannot be used together, if both are set to True, the cache will take precedence and the request will be awaited.
        """
        if cache:
            d_response = await self._read_cache(id=id, method=method)
        else:
            d_response = await self._read(id=id, method=method, as_task=as_task)

        return d_response

    async def _read(
            self, id: Optional[Union[int, str]] = None, method: str = "search", renew_cache: bool = False,
            as_task: bool = False
    ) -> DirectusResponse:
        """
        Send query to server.
        """

        method = "get" if id is not None else method
        if method == "search":
            response = self.directus.connection.request(
                "search", self.uri,
                json={"query": self.params},
                auth=self.directus.auth
            )
        elif method == "get":
            url = f"{self.uri}/{id}" if id is not None else self.uri
            response = self.directus.connection.get(url, params=self.params, auth=self.directus.auth)
        else:
            raise ValueError(f"Method '{method}' not supported")

        d_response = DirectusResponse(response, query=self.params, collection=self.collection_class)

        if as_task:
            self.directus.tasks.append(d_response)
        else:
            await d_response.gather_response()

        # Check for existing cache and renew it
        if not renew_cache:
            async with self._lock_2:
                query_key_str = self._get_query_string_key()

                # Try to find query in cache
                cached_response = await self.directus.cache.get(query_key_str)

                if cached_response:
                    # Renew the cache value
                    await self.directus.cache.add(query_key_str, d_response.to_json())

        return d_response

    async def _read_cache(
            self, id: Optional[Union[int, str]] = None, method: str = "search"
    ) -> DirectusResponse:
        """
        Get response from cache.
        """
        async with self._lock:
            query_key_str = self._get_query_string_key()

            # Try to find query in cache
            cached_response = await self.directus.cache.get(query_key_str)

            if cached_response:
                print("FROM CACHE")
                d_response = DirectusResponse.from_json(cached_response)
            else:
                print(self.directus.cache._cache)
                print("FROM NEW")
                d_response = await self._read(id=id, method=method, renew_cache=True)

                # Add results to cache
                await self.directus.cache.add(query_key_str, d_response.to_json())

            return d_response

    async def clear_cache(self):
        query_key_str = self._get_query_string_key()

        # Try to find query in cache
        d_res = await self.directus.cache.delete(query_key_str)
        return d_res

    def _get_query_string_key(self):
        """
        Generate request key for cache.
        """

        query_str = jsonlib.dumps(self.params)

        return f"{self.collection}_{query_str}"

    async def subscribe(self, uri: str, event_type: Optional[str] = None, uid: Optional[str] = None) -> Tuple[
        'Data', 'WebSocketClientProtocol']:
        """
        Returns authentication confirmation message and the client websocket.
        """
        ws = await websockets.connect(uri)

        # Authentication
        auth_data = jsonlib.dumps({
            "type": "auth",
            "access_token": self.directus._token
        })

        await ws.send(auth_data)
        auth_res = await ws.recv()

        # Subscription
        subsc_data = {
            "type": "subscribe",
            "collection": self.collection,
            "query": self.params
        }

        if event_type:
            subsc_data['event'] = event_type

        if uid:
            subsc_data['uid'] = uid

        await ws.send(jsonlib.dumps(subsc_data))
        subsc_res = await ws.recv()

        return auth_res, ws

    async def create(self, items: Union[dict, List[dict]]) -> DirectusResponse:
        assert isinstance(items, (dict, list))

        response = await self.directus.connection.post(self.uri, json=items, auth=self.directus.auth)
        return DirectusResponse(response, collection=self.collection_class)

    @overload
    async def update(self, ids: Optional[Union[int, str]], items: dict) -> DirectusResponse:
        ...

    @overload
    async def update(self, ids: List[Union[int, str]], items: list) -> DirectusResponse:
        ...

    async def update(self, ids, items):
        if isinstance(ids, (int, str, None)) and isinstance(items, dict):
            if ids is None:
                response = await self.directus.connection.patch(self.uri, json=items, auth=self.directus.auth)
            else:
                response = await self.directus.connection.patch(f"{self.uri}/{ids}", json=items,
                                                                auth=self.directus.auth)
            return DirectusResponse(response, collection=self.collection_class)
        elif isinstance(ids, list) and isinstance(items, list):
            payload = {
                "keys": ids,
                "data": items
            }
            response = await self.directus.connection.patch(self.uri, json=payload, auth=self.directus.auth)
            return DirectusResponse(response, collection=self.collection_class)

        raise TypeError(
            f"This method supports the following argument pairs: \n"
            f"ids: int | str | None, items: dict\n"
            f"ids: list[int | str], items: list\n"
            f"You provided: ids={type(ids)}, items={type(items)}"
        )

    async def delete(self, ids: Union[int, str, List[Union[int, str]]]) -> DirectusResponse:
        if isinstance(ids, (int, str)):
            response = await self.directus.connection.delete(f'{self.uri}/{ids}', auth=self.directus.auth)
            return DirectusResponse(response, collection=self.collection_class)
        elif isinstance(ids, list):
            response = await self.directus.connection.delete(self.uri, json=ids, auth=self.directus.auth)
            return DirectusResponse(response, collection=self.collection_class)

        raise TypeError(
            f"The ids argument must be one of the following types: int | str | list[int | str]\n"
            f"You provided: {ids}"
        )



import json
from typing import Union, List

from py_directus.expression import Expression
from py_directus.operators import AGGREGATION_OPERATORS, AggregationOperators


class Agg(Expression):
    """
    Aggregation field.
    """

    def __init__(self, *args, **kwargs):
        # Initialize the query as an empty dictionary
        self.query = {}

        # Arguments
        for arg in args:
            if isinstance(arg, str):
                self._from_string(arg)
        
        # Operator
        if "operator" in kwargs or "fields" in kwargs:
            kwrg_operator = kwargs.pop("operator", None)
            kwrg_fields = kwargs.pop("fields", None)

            self._from_operator(kwrg_operator, kwrg_fields)

        # Keyword Arguments
        for key, value in kwargs.items():
            self._from_kwarg(key, value)

        if not self.query:
            self.query["count"] = "*"

    def _from_operator(self, operator: Union[AggregationOperators, None] = AggregationOperators.Count, fields: Union[str, List[str], None] = "*"):
        # Clean arguments
        if isinstance(fields, list):
            cln_fields = [Agg.parse_key(fld)[0] for fld in fields]
        else:
            cln_fields = Agg.parse_key(fields)[0]

        # Add given arguments
        self._add(operator, cln_fields)

    def _from_string(self, field_def: str):
        # Clean arguments
        field, operator = Agg.parse_key(field_def)

        # Add given arguments
        self._add(operator, field)
    
    def _from_kwarg(self, operator: str, fields: Union[str, List[str]]):
        # Clean arguments
        if isinstance(fields, list):
            cln_fields = [Agg.parse_key(fld)[0] for fld in fields]
        else:
            cln_fields = Agg.parse_key(fields)[0]

        # Add given arguments
        self._add(operator, cln_fields)

    @staticmethod
    def parse_key(key):
        field = None
        operator = None

        for _operator in AGGREGATION_OPERATORS:
            if key.endswith("_" + _operator):
                field = key[:-len("_" + _operator)]
                operator = _operator
                break

        # No operator matched, which we assume was not provided.
        # Thus we use the default 'count' operator.
        if field is None:
            operator = "count"
            field = key.replace("__", ".")
            return field, operator

        # Deep fields
        field = field.replace("__", ".")

        if operator:
            operator = operator[1:]

        if field == "":
            field = None
        return field, operator

    def _add(self, operator: Union[AggregationOperators, str, None] = None, fields: Union[str, List[str], None] = None):
        if operator and fields:
            if isinstance(operator, AggregationOperators):
                needed_operator = operator.value
            else:
                needed_operator = operator

            if needed_operator in self.query:
                orig_value = self.query[needed_operator]

                # Encapsulate the value in a list
                if not isinstance(orig_value, list):
                    orig_value = [orig_value]

                # Add fields to value list
                if isinstance(fields, list):
                    orig_value = [*orig_value, *fields]
                else:
                    orig_value.append(fields)

                self.query[needed_operator] = orig_value
            else:
                self.query[needed_operator] = fields

    def __add__(self, other: 'Agg'):
        # Implement addition logic
        for operator, fields in other.query.items():
            self._add(operator, fields)
        return self

    def __str__(self):
        return f"\n{json.dumps(self.query, indent=2)}\n"

    def __json__(self):
        return json.dumps(self.query, indent=2)



from datetime import datetime
from typing import Union, Tuple

try:
    import human_readable
except ImportError:
    human_readable = None

from typing_extensions import Annotated

from pydantic.functional_validators import AfterValidator


class ModelDateTime(datetime):
    """
    """

    DATE_FORMAT: str = "%d/%m/%Y"
    TIME_FORMAT: Tuple[str, str] = ("%H:%M:%S", "%H:%M")
    DATETIME_FORMAT: Tuple[str, str] = ("%d/%m/%Y %H:%M:%S", "%d/%m/%Y %H:%M")

    # Pydantic field data type
    field = Annotated[
        datetime,
        AfterValidator(lambda v: ModelDateTime.fromisoformat(v.isoformat())),
    ]

    def str_only_date(self):
        return self.strftime(self.DATE_FORMAT)

    def str_only_time(self, with_seconds: bool=False):
        if with_seconds:
            return self.strftime(self.TIME_FORMAT[0])
        return self.strftime(self.TIME_FORMAT[1])

    def str_date_time(self, with_seconds: bool=False):
        if with_seconds:
            return self.strftime(self.DATETIME_FORMAT[0])
        return self.strftime(self.DATETIME_FORMAT[1])

    if human_readable:
        def str_human_readable(self):
            return human_readable.date_time(ModelDateTime.now() - self)



import re
import inspect

from pydantic._internal._config import ConfigWrapper
from pydantic.main import _model_construction, BaseModel


class DirectusModelMetaclass(_model_construction.ModelMetaclass):
    """
    """

    def __new__(cls, name, bases, namespace, *args, **kwargs):
        """ Perform check for `collection` declaration """

        tmp_config_wrapper = ConfigWrapper.for_model((), namespace, kwargs)
        declared_collection = tmp_config_wrapper.config_dict.get("collection", None)

        completed_class = super().__new__(cls, name, bases, namespace, *args, **kwargs)

        class_config = getattr(completed_class, "model_config", None)

        if not declared_collection:
            output = re.findall("[A-Z]?[a-z]+", name)
            class_config["collection"] = "_".join([word.lower() for word in output])

        return completed_class


class DirectusModel(BaseModel, metaclass=DirectusModelMetaclass):
    pass



import functools
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Union, Optional, Type

from . import globals as glob_vars

if TYPE_CHECKING:
    from fastapi import FastAPI


@asynccontextmanager
async def _lifespan(app: Union['FastAPI', None] = None, directus_base_url: str = None, directus_admin_token: str = None, *args, **kwargs):
    """
    Only when used directly as context manager we have access to the `FastAPI` instance through the `app` argument.
    """

    # Initialize global clients
    await glob_vars.async_init(directus_base_url, directus_admin_token)

    try:
        yield
    finally:
        # Close global clients
        if glob_vars.directus_admin:
            await glob_vars.directus_admin.logout()
        if glob_vars.directus_public:
            await glob_vars.directus_public.logout()
        if glob_vars.directus_session:
            await glob_vars.directus_session.aclose()


def lifespan(*og_args, **og_kwargs):
    def wrapper(cm=None, *ag_args, **ag_kwargs):
        if callable(cm):
            @asynccontextmanager
            @functools.wraps(cm)
            async def wrapped(*args, **kwargs):
                # NOTE: FIND A WAY TO PROVIDE THE `app` ARGUMENT TO OUR LIFESPAN
                async with _lifespan(*og_args, **og_kwargs) as og_res:
                    async with cm(*args, **kwargs) as cm_res:
                        # Occupy resources
                        try:
                            yield cm_res
                        finally:
                            # Release resources
                            pass
            return wrapped
        else:
            return _lifespan(*og_args, *ag_args, **og_kwargs, **ag_kwargs)
    return wrapper


def init_directus(app: 'FastAPI', directus_base_url: str, directus_admin_token: str):
    """
    Wrap the lifespan context manager of FastAPI with our own.
    """

    cm = app.router.lifespan_context
    app.router.lifespan_context = lifespan(
        directus_base_url=directus_base_url, 
        directus_admin_token=directus_admin_token
    )(cm)



from enum import Enum


FILTER_OPERATORS = {
    "_eq": "Equals",
    "_neq": "Doesn't equal",
    "_lt": "Less than",
    "_lte": "Less than or equal to",
    "_gt": "Greater than",
    "_gte": "Greater than or equal to",
    "_in": "Is one of",
    "_nin": "Is not one of",
    "_null": "Is null",
    "_nnull": "Isn't null",
    "_contains": "Contains",
    "_icontains": "Contains case-insensitive",
    "_ncontains": "Doesn't contain",
    "_starts_with": "Starts with",
    "_istarts_with": "Starts with case-insensitive",
    "_nstarts_with": "Doesn't start with",
    "_nistarts_with": "Doesn't start with case-insensitive",
    "_ends_with": "Ends with",
    "_iends_with": "Ends with case-insensitive",
    "_nends_with": "Doesn't end with",
    "_niends_with": "Doesn't end with case-insensitive",
    "_between": "Is between",
    "_nbetween": "Isn't between",
    "_empty": "Is empty",
    "_nempty": "Isn't empty",
    "_intersects": "Intersects",
    "_nintersects": "Doesn't intersect",
    "_intersects_bbox": "Intersects Bounding box",
    "_nintersects_bbox": "Doesn't intersect bounding box",
    "_and": "AND",
    "_or": "OR",
}


class AggregationOperators(str, Enum):
    Count = "count"
    CountDistinct = "countDistinct"
    CountAll = "countAll"
    Sum = "sum"
    SumDistinct = "sumDistinct"
    Average = "avg"
    AverageDistinct = "avgDistinct"
    Minimum = "min"
    Maximum = "max"

AGGREGATION_OPERATORS = {
    "_count": AggregationOperators.Count,
    "_countDistinct": AggregationOperators.CountDistinct,
    "_countAll": AggregationOperators.CountAll,
    "_sum": AggregationOperators.Sum,
    "_sumDistinct": AggregationOperators.SumDistinct,
    "_avg": AggregationOperators.Average,
    "_avgDistinct": AggregationOperators.AverageDistinct,
    "_min": AggregationOperators.Minimum,
    "_max": AggregationOperators.Maximum
}



from http import HTTPStatus


class ApiException(Exception):
    def __init__(self, message: str = "", extras: dict = None, status_code: HTTPStatus = HTTPStatus.NOT_ACCEPTABLE):
        self.status_code = status_code
        self.message = message
        self.extras = extras

    def __str__(self):
        return self.message



import asyncio
from functools import wraps
from http import HTTPStatus
from typing import Union, Optional, List

from dotenv import load_dotenv
from httpx import AsyncClient

from fastapi import HTTPException
from fastapi.security.utils import get_authorization_scheme_param

from starlette.requests import Request
from starlette.responses import Response
from starlette.status import HTTP_307_TEMPORARY_REDIRECT

from py_directus import Directus

from .utils import (
    get_directus_from_token, 
    role_to_id
)
from .exceptions import ApiException


class HeaderAndCookieBearer:
    """
    Creates a `Directus` instance from provided credentials in request headers.
    """

    @staticmethod
    async def check_header(request: Request) -> Optional[Directus]:
        header_authorization = request.headers.get("Authorization")
        scheme, credentials = get_authorization_scheme_param(header_authorization)
        if not (header_authorization and scheme and credentials) or scheme.lower() != "bearer":
            return None
        return await get_directus_from_token(credentials)

    @staticmethod
    async def check_cookie(request: Request) -> Optional[Directus]:
        access_token: str = request.cookies.get("access_token")
        refresh_token: str = request.cookies.get("refresh_token")
        return await get_directus_from_token(access_token, refresh_token) if (access_token and refresh_token) else None

    async def __call__(self, request: Request, response: Response) -> Optional[Directus]:
        header_directus = await HeaderAndCookieBearer.check_header(request)
        cookie_directus = await HeaderAndCookieBearer.check_cookie(request)
        directus = header_directus or cookie_directus

        if directus:
            return directus

        raise HTTPException(
            status_code=HTTP_307_TEMPORARY_REDIRECT,
            headers={'Location': '/login'}
        )


directus_auth = HeaderAndCookieBearer()


def assert_role(allowed_roles: Union[str, List[str]] = None):
    def assert_role_inner(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            _allowed_roles = allowed_roles
            if isinstance(_allowed_roles, str):
                _allowed_roles = [_allowed_roles]

            await role_to_id
            roles_ids = [role_to_id(role) for role in _allowed_roles]
            roles_ids = [item for sublist in roles_ids for item in sublist]  # flatten list
            directus = None

            if "directus" in kwargs:
                directus = kwargs['directus']
            else:
                for arg in args:
                    if isinstance(arg, Directus):
                        directus = arg
                        break

            if not directus:
                raise ApiException(
                    "Not allowed, you must be authenticated", 
                    {'not_allowed': _allowed_roles},
                    HTTPStatus.FORBIDDEN
                )

            directus_user = await directus.user
            user_role = directus_user.role

            if user_role not in roles_ids:
                raise ApiException('Not allowed', {'not_allowed': _allowed_roles}, HTTPStatus.FORBIDDEN)
            if asyncio.iscoroutinefunction(func):  # check if wrapped function is a coroutine function
                return await func(*args, **kwargs)
            else:
                return func(*args, **kwargs)

        return wrapper

    return assert_role_inner



import asyncio
from enum import Enum
from typing import Union, Optional, List

from py_directus.models import Role
from py_directus import Directus
from py_directus.fast_api import globals


async def get_directus_login(email: str, password: str) -> Directus:
    d = await Directus(globals.directus_url, connection=globals.directus_session, email=email, password=password)
    globals.cached_directus_instances[d.token] = d
    return d


async def get_directus_from_token(access_token, refresh_token=None) -> Optional[Directus]:
    directus = await Directus(globals.directus_url, token=access_token, refresh_token=refresh_token,
                              connection=globals.directus_session)
    await directus.user  # noqa
    return directus


async def directus_logout(directus: Directus):
    globals.cached_directus_instances.pop(directus.token, None)
    await directus.logout()


class Roles(str, Enum):
    ADMIN = "Administrator"
    COMPANY = "Company"
    LICENSE = "License"
    CREDIT = "Credit"
    DEVICE = "Device"
    PARTNER = "Partner"


class RoleToID:
    def __init__(self):
        self.roles: Optional[List[Role]] = None

    def __await__(self):
        async def closure():
            if getattr(self, "roles", None) is None:
                # Perform login manually, because the global was instantiated without awaiting
                await globals.directus_admin
                roles = await globals.directus_admin.collection(Role).read()
                self.roles = {role.name: role.id for role in roles.items}
            return self

        return closure().__await__()

    def __call__(self, role: Union[str, Roles]) -> [str]:  # noqa
        if isinstance(role, Roles):
            role = role.value
        return [self.roles[role]]


role_to_id = RoleToID()



import json

from rich import print  # noqa
from rich.console import Console  # noqa

from py_directus.expression import Expression
from py_directus.operators import FILTER_OPERATORS


class F(Expression):
    """
    Filter field.
    """

    def __init__(self, **kwargs):
        # Initialize the query as an empty dictionary
        self.query = {}

        # Parse keyword arguments into query format
        for key, value in kwargs.items():
            field, operator = F.parse_key(key)
            if not field:  # its a logical operator
                self.query[operator] = value
                continue
            field_parts = field.split(".")
            field = field_parts[0]

            if field not in self.query:
                self.query[field] = {}
            # convert deep fields from dot notation to nested dicts
            if len(field_parts) > 1:
                inner_dict = {operator: value}
                for part in reversed(field_parts[1:]):
                    inner_dict = {part: inner_dict}
                self.query[field] = {**self.query[field], **inner_dict}
            else:
                self.query[field][operator] = value

        if len(kwargs.items()) > 1:
            inner_queries = []
            for key, value in kwargs.items():
                inner_queries.append(F(**{key: value}).query)
            self.query = {"_and": inner_queries}

    @staticmethod
    def parse_key(key):
        field = None
        operator = None

        for _operator in FILTER_OPERATORS:
            if key.endswith("_" + _operator):
                field = key[:-len("_" + _operator)]
                operator = _operator
                break

        # No operator matched, which we assume was not provided.
        # Thus we use the default 'equals' operator.
        if field is None:
            operator = "_eq"
            field = key.replace("__", ".")
            return field, operator

        # Deep fields
        field = field.replace("__", ".")

        if field == "":
            field = None
        return field, operator

    def combine(self, other: 'F', operator: str) -> 'F':
        if other is None:
            return self
        else:
            if other.query == {}:
                return self
            elif self.query == {}:
                return other
            return F(**{operator: [self.query, other.query]})

    def __and__(self, other: 'F'):
        return self.combine(other, "__and")

    def __or__(self, other: 'F'):
        return self.combine(other, "__or")

    def __rand__(self, other: 'F'):
        return self.__and__(other)

    def __ror__(self, other: 'F'):
        return self.__or__(other)

    def __str__(self):
        return f"\n{json.dumps(self.query, indent=2)}\n"

    def __json__(self):
        return json.dumps(self.query, indent=2)

    def convert_query_to_string(self, input_dict):
        if "_or" in input_dict or "_and" in input_dict:
            key = "_or" if "_or" in input_dict else "_and"
            op_list = input_dict[key]
            return "(" + f"{FILTER_OPERATORS[key]}".join(
                [self.convert_query_to_string(subdict) for subdict in op_list]) + ")"
        else:
            key, value = list(input_dict.items())[0]
            if isinstance(value, dict):
                inner_str = " ".join([f"{FILTER_OPERATORS[k]} {v}" for k, v in value.items()])

                return f"({key} {inner_str})"

    def get_explanation(self, tab_char="  "):
        input_ex = self.convert_query_to_string(self.query)
        tabs = 0
        index = 0
        min_tabs = 500
        result = ""

        while True:
            if input_ex[index] == '(':
                tabs += 1
                result += '\n' + tab_char * tabs + input_ex[index]
            elif input_ex[index] == ')':
                tabs -= 1
                result += '\n' + tab_char * tabs + input_ex[index]
            else:
                min_tabs = min(min_tabs, tabs)
                result += input_ex[index]
            index += 1
            if index == len(input_ex):
                break

        result = result.replace('(', '').replace(')', '')
        result = result.split('\n')
        result = filter(lambda x: x.strip(tab_char) != '', result)
        result = list(map(lambda x: x[min_tabs * len(tab_char):], result))

        return "\n".join(result)

    def print_explanation(self, tab_char="  "):
        console = Console()
        explain = self.get_explanation(tab_char)

        for i in explain.split('\n'):
            if "AND" in i or "OR" in i:
                console.print(i, style="bold")
                continue
            count_tabs = i.count(tab_char)
            console.print(i, style=f"color({count_tabs})", highlight=False)



from .directus import Directus
from .filter import F

try:
    from .fast_api.auth import HeaderAndCookieBearer
    from .fast_api import globals
    from .fast_api.globals import directus_admin, directus_session, directus_auth
    from .fast_api.lifespan import init_directus
except ImportError:
    pass



import os
import pathlib
import errno
import platformdirs
import aiofiles

from py_directus.utils import get_random_string


class SuspiciousFileOperation(Exception):
    """
    A Suspicious filesystem operation was attempted.
    
    Origin: https://github.com/django/django/blob/3.2.23/django/core/files/storage.py
    """
    pass


def _validate_file_name(name, allow_relative_path=False):
    """
    Origin: https://github.com/django/django/blob/3.2.23/django/core/files/utils.py
    """

    # Remove potentially dangerous names
    if os.path.basename(name) in {'', '.', '..'}:
        raise SuspiciousFileOperation(f"Could not derive file name from '{name}'")

    if allow_relative_path:
        # Use PurePosixPath() because this branch is checked only in
        # FileField.generate_filename() where all file paths are expected to be
        # Unix style (with forward slashes).
        path = pathlib.PurePosixPath(name)
        if path.is_absolute() or ".." in path.parts:
            raise SuspiciousFileOperation(
                f"Detected path traversal attempt in '{name}'"
            )
    elif name != os.path.basename(name):
        raise SuspiciousFileOperation(f"File name '{name}' includes path elements")

    return name


def _get_alternative_name(file_root, file_ext):
    """
    Return an alternative filename, by adding an underscore and a random 7
    character alphanumeric string (before the file extension, if one
    exists) to the filename.

    Origin: https://github.com/django/django/blob/3.2.23/django/core/files/storage.py
    """
    return f"{file_root}_{get_random_string(7)}{file_ext}"


def _get_available_name(name, max_length=None):
    """
    Return a filename that's free on the target storage system and
    available for new content to be written to.

    Origin: https://github.com/django/django/blob/3.2.23/django/core/files/storage.py
    """

    name = str(name).replace('\\', '/')
    dir_name, file_name = os.path.split(name)

    if ".." in pathlib.PurePath(dir_name).parts:
        raise SuspiciousFileOperation(f"Detected path traversal attempt in '{dir_name}'")

    _validate_file_name(file_name)
    file_root, file_ext = os.path.splitext(file_name)

    # If the filename already exists, generate an alternative filename
    # until it doesn't exist.
    # Truncate original name if required, so the new filename does not
    # exceed the max_length.
    while os.path.exists(_get_default_downloads_path(name)) or (max_length and len(name) > max_length):
        # file_ext includes the dot.
        name = os.path.join(dir_name, _get_alternative_name(file_root, file_ext))

        if max_length is None:
            continue

        # Truncate file_root if max_length exceeded.
        truncation = len(name) - max_length

        if truncation > 0:
            file_root = file_root[:-truncation]

            # Entire file_root was truncated in attempt to find an available filename.
            if not file_root:
                raise SuspiciousFileOperation(
                    f"Storage can not find an available filename for '{name}'. "
                    f"Please make sure that you allows sufficient 'max_length'."
                )
            name = os.path.join(dir_name, _get_alternative_name(file_root, file_ext))

    return name


def _get_default_downloads_path(filename: str):
    """
    """

    path = os.path.join(platformdirs.user_downloads_dir(), "py_directus")

    # Make directories
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise

    # Append the file name
    return os.path.join(path, filename)


async def save_file(filename: str, content: bytes) -> str:
    """
    """

    name = _get_available_name(filename)
    full_path = _get_default_downloads_path(name)

    # Create file from received data

    while True:
        try:
            # Open for exclusive creation (as bytes)
            async with aiofiles.open(full_path, 'xb') as f:
                await f.write(content)
        except FileExistsError:
            # A new name is needed if the file exists.
            name = _get_available_name(name)
            full_path = _get_default_downloads_path(name)
        else:
            # Everything went OK, so get out of the infinite loop
            break

    # Ensure that the name returned from the storage system is still valid.
    _validate_file_name(name, allow_relative_path=True)

    return name



#!/usr/bin/env python

from pathlib import Path

from setuptools import setup, find_packages

with open('requirements.txt') as requirements_file:
    requirements = requirements_file.readlines()

with open('fastapi_requirements.txt') as fastapi_requirements_file:
    fastapi_requirements = fastapi_requirements_file.readlines()

readme = Path('README.md').read_text()
version = '0.0.17'
if "VERSION_PLACEHOLDER" in version:
    version = '0.0.8'
setup(
    author="Panos Stavrianos",
    author_email='panos@orbitsystems.gr',
    python_requires='>=3.6',
    description="Python wrapper for asynchronous interaction with Directus",
    long_description=readme,
    long_description_content_type='text/markdown',
    install_requires=requirements,
    extras_require={
        "FastAPI": fastapi_requirements,
    },
    license="MIT license",
    include_package_data=True,
    keywords=['python', 'directus', 'async', 'asyncio', 'api', 'wrapper'],
    name='py_directus',
    packages=find_packages(include=['py_directus', 'py_directus.*']),
    url='https://github.com/panos-stavrianos/py-directus',
    version=version,
    zip_safe=False
)


# Based on the cache implementation in Zeep: https://github.com/mvantellingen/python-zeep/blob/4.2.1/src/zeep/cache.py
import base64
import logging
from abc import ABC, abstractmethod
from typing import Dict, Tuple, Union, Optional
from datetime import datetime, timedelta, timezone

from py_directus.utils import get_random_string


logger = logging.getLogger(__name__)


class Base(ABC):
    """
    Base class for caching backends.
    """

    @abstractmethod
    async def add(self, query, content):
        raise NotImplementedError()

    @abstractmethod
    async def get(self, query):
        raise NotImplementedError()

    @abstractmethod
    async def delete(self, query):
        raise NotImplementedError()

    @abstractmethod
    async def clear(self, select_all):
        raise NotImplementedError()


class SimpleMemoryCache(Base):
    """
    Simple in-memory caching using dict lookup with support for timeouts
    """

    # cache persistent throughout class instances, thread-safe by default
    _cache: Dict[str, Tuple[datetime, Union[str, bytes]]] = {}

    def __init__(self, unique_id: str, timeout: int=3600):
        self._timeout = timeout

        self.unique_id = unique_id

    async def add(self, query: str, content: Union[str, bytes]):
        q_key = self._get_query_key(query)

        logger.debug("Caching contents of %s", q_key)

        if not isinstance(content, (str, bytes)):
            raise TypeError(
                "a bytes-like object is required, not {}".format(type(content).__name__)
            )

        self._cache[q_key] = (datetime.utcnow(), content)
        return True

    async def get(self, query: str):
        q_key = self._get_query_key(query)

        try:
            created, content = self._cache[q_key]
        except KeyError:
            pass
        else:
            if _is_expired(created, self._timeout):
                self._delete(q_key)
            else:
                logger.debug("Cache HIT for %s", q_key)

                return content

        logger.debug(f"Cache MISS for {q_key}")

        return None

    async def delete(self, query: str):
        q_key = self._get_query_key(query)

        d_res = self._delete(q_key)

        logger.debug(f"Deleted contents of {q_key}")

        return d_res

    async def clear(self, select_all: bool=False):
        if select_all:
            self._cache.clear()
        else:
            for key in list(self._cache):
                if key.startswith(self.unique_id):
                    self._delete(key)
        
        logger.debug(f"Cleared cache ({select_all})")

        return True

    def _delete(self, q_key: str):
        if self._cache.pop(q_key, None) is not None:
            return True
        return False

    # NOTE: Deprecated
    # def _get_unique_id(self) -> str:
    #     """
    #     Expose the version prefix to be used in content serialization.
    #     """
    #     length = 25
    #     retries = 25

    #     is_unique = False
    #     unique_id = None

    #     while retries > 0:
    #         unique_id = get_random_string(length)

    #         if unique_id not in self._cache:
    #             is_unique = True
    #             break

    #         retries -= 1

    #     # In case there was no spare id
    #     if not is_unique:
    #         raise Exception(f"No unique id could be found within the specified {retries} retries")

    #     return unique_id

    def _get_query_key(self, query: str) -> str:
        """
        Expose the version prefix to be used in content serialization.
        """
        q_key = base64.b64encode(query.encode('utf-8'))

        prefix = f"{self.unique_id}:{q_key.decode('utf-8')}"
        return prefix


def _is_expired(value: datetime, timeout: int) -> bool:
    """
    Return boolean if the value is expired

    timeout: Number of seconds after which the value is considered expired
    """
    if timeout is None:
        return False

    now = datetime.utcnow().replace(tzinfo=timezone.utc)
    max_age = value.replace(tzinfo=timezone.utc)
    max_age += timedelta(seconds=timeout)
    return now > max_age



import asyncio
import datetime
import inspect
import os
import re
# import aiofiles
from typing import Union, Optional, Type

import magic
from httpx import AsyncClient, Auth, Response
from pydantic import BaseModel

from py_directus.cache import SimpleMemoryCache
from py_directus.directus_request import DirectusRequest
from py_directus.directus_response import DirectusResponse
from py_directus.models import Role, User
from py_directus.storage import save_file
from py_directus.transformation import ImageFileTransform
from py_directus.utils import parse_translations


class BearerAuth(Auth):
    def __init__(self, token: str):
        self.token = token

    def auth_flow(self, request):
        if self.token is not None:
            request.headers["Authorization"] = f"Bearer {self.token}"
        yield request


class Directus:
    """
    Asynchronous Client for Directus API communication.
    """

    def __init__(
            self, url: str, email: str = None, password: str = None, token: str = None, refresh_token: str = None,
            connection: AsyncClient = None, user_model: Type[User] = User
    ):
        self.expires = None
        self.expiration_time = None
        self.refresh_token = refresh_token
        self.url: str = url

        # Credentials
        self._token: Optional[str] = None
        self._user: Optional[User] = None
        self._user_model: Type[User] = user_model

        self.email = email
        self.password = password

        self.static_token = token
        self.token = token or None

        # Connection
        self.connection = connection or AsyncClient()
        self.auth = BearerAuth(self._token)

        # Cache
        self.cache: SimpleMemoryCache = None

        # Any async tasks for later gathering
        self.tasks: list[DirectusResponse] = []

    async def gather(self):
        """
        Gather all async tasks.
        """
        print("Gathering tasks", self.tasks)
        await asyncio.gather(*[task.gather_response() for task in self.tasks])
        self.tasks.clear()

    def __await__(self):
        async def closure():
            # Perform login when credentials are present and no token
            if self.email and self.password and not self._token:
                await self.login()
            if not self.cache:
                self.cache = SimpleMemoryCache(self._token)
            return self

        return closure().__await__()

    def collection(self, collection: Union[Type[BaseModel], str]) -> DirectusRequest:
        """
        Set collection to be used.
        """

        if inspect.isclass(collection):
            assert issubclass(collection, BaseModel), (
                f"The provided collection model must be a subclass of pydantic.BaseModel"
            )

            collection_name = collection.model_config.get("collection", None)

            assert collection_name is not None

            return DirectusRequest(self, collection_name, collection)
        elif isinstance(collection, str):
            return DirectusRequest(self, collection, None)

        raise TypeError(
            f"The `collection` argument must be either a string or a subclass of the pydantic.BaseModel class.\n"
            f"You gave: {collection}"
        )

    async def me(self, user_model: Union[Type[User], str, None] = None) -> DirectusResponse:
        """
        Retrieve logged in user's information.
        """

        if not user_model:
            user_model = self._user_model

        response_obj = await self.collection(user_model).read("me")

        return response_obj

    async def roles(self) -> DirectusResponse:
        """
        Retrieve list of roles in system.
        """

        response_obj = await self.collection(Role).read()
        return response_obj

    async def read_settings(self) -> DirectusResponse:
        response_obj = await DirectusRequest(self, "directus_settings").read(method='get')
        return response_obj

    async def update_settings(self, data) -> DirectusResponse:
        response_obj = await DirectusRequest(self, "directus_settings").update(None, data)
        return response_obj

    async def read_translations(self) -> dict[str, dict[str, str]]:
        """
        NOTE: TO BE REDESIGNED
        """
        items = await self.collection("translations").fields(
            'key', 'translations.languages_code', 'translations.translation'
        ).read().items

        return parse_translations(items)

    async def create_translations(self, keys: list[str]):
        """
        NOTE: TO BE REDESIGNED
        """
        response_obj = await self.collection("translations").create([{"key": key} for key in keys])
        return response_obj

    async def download_file(
            self, file_id: str,
            fit: Optional[str] = None,
            width: Optional[int] = None, height: Optional[int] = None,
            quality: Optional[int] = None,
            withoutEnlargement: Optional[bool] = None,
            img_format: Optional[str] = None,
            **kwargs
    ) -> Response:
        url = f"{self.url}/assets/{file_id}"

        request_params = {
            "download": ""
        }

        # Image transformation parameters
        img_transform_parameters = ImageFileTransform(
            fit=fit,
            width=width,
            quality=quality,
            withoutEnlargement=withoutEnlargement,
            img_format=img_format,
            **kwargs
        ).parameters

        if img_transform_parameters:
            request_params.update(img_transform_parameters)

        response = await self.connection.get(url, params=request_params)

        if response.status_code == 200:
            # Get file name
            d = response.headers['content-disposition']
            fname = re.findall("filename=[\"\'](.+)[\"\']", d)
            cln_fname = fname[0] if fname else file_id

            await save_file(cln_fname, response.content)

        return response

    async def upload_file(self, file_path: str) -> Response:
        url = f"{self.url}/files"

        # file name with extension
        file_name = os.path.basename(file_path)
        file_mime = magic.from_file(file_path, mime=True)

        data = {
            "title": os.path.splitext(file_name)[0],
            # "folder": "foreign_key"
        }

        try:
            # File
            # NOTE: CANNOT USE ASYNCHRONOUS FILES BECAUSE THEY ARE NOT SUPPORTED BY HTTPX
            # https://github.com/encode/httpx/issues/1620
            # f = await aiofiles.open(file_path, 'rb')
            f = open(file_path, 'rb')

            files = {
                "file": (file_name, f, file_mime)
            }

            response = await self.connection.post(url, data=data, files=files, auth=self.auth)
        finally:
            f.close()

        return response

    async def clear_cache(self, clear_all: bool = False):
        """
        Clear cache:

        :param clear_all: If set to `True`, absolutely ALL records will be deleted.
        """
        clr_res = await self.cache.clear(clear_all)

        return clr_res

    async def __aenter__(self):
        return self

    @property
    def token(self):
        return self._token

    @token.setter
    def token(self, token):
        self._token = token
        self.auth = BearerAuth(self._token)

    @property
    async def user(self):
        if self._user is None:
            user = await self.me()
            self._user = user.item
        return self._user

    async def auth_request(self, endpoint, payload):
        url = f"{self.url}/{endpoint}"

        r = await self.connection.post(url, json=payload)
        response = DirectusResponse(r).item

        self._token = response['access_token']
        self.refresh_token = response['refresh_token']
        self.expires = response['expires']  # in milliseconds
        self.expiration_time: datetime.datetime = (
                datetime.datetime.now() + datetime.timedelta(milliseconds=self.expires)
        )
        self.auth = BearerAuth(self._token)

    async def login(self):
        endpoint = "auth/login"

        if self.static_token:
            self._token = self.static_token
            return

        payload = {
            'email': self.email,
            'password': self.password
        }
        await self.auth_request(endpoint, payload)

    async def refresh(self):
        endpoint = "auth/refresh"
        payload = {
            'refresh_token': self.refresh_token,
            "mode": "json"
        }
        await self.auth_request(endpoint, payload)

    async def logout(self) -> bool:
        url = f"{self.url}/auth/logout"
        response = await self.connection.post(url)

        self.connection.auth = None
        self._token = None
        self.refresh_token = None
        self.expires = None
        self.expiration_time = None

        # Clear cache
        if self.cache:
            await self.cache.clear()

        return response.status_code == 200

    async def close_connection(self):
        await self.connection.aclose()

    async def __aexit__(self, *args):
        print("Closing connection")
        # Exception handling here
        await self.logout()
        await self.close_connection()



from __future__ import annotations

import inspect
import json as jsonlib
from typing import Any, Union, TypeVar, List, Coroutine

try:
    from rich import print  # noqa
    from rich.console import Console  # noqa
except:
    pass

from httpx import Response

from pydantic import BaseModel, TypeAdapter


class DirectusResponse:
    T = TypeVar("T", bound=BaseModel)

    def __init__(self, response: Union[Response, Coroutine, None] = None, query: dict = None, collection: Any = None):
        self.response: Union[Response, Coroutine, None] = response
        self.response_status: Union[int, None] = None
        self.query: dict = query
        self.collection: Any = collection
        self.json: Union[dict, None] = None

        self.parse_response()

    def parse_response(self):
        # In case we were given a response, then it is not a cache response
        if self.response and not inspect.iscoroutine(self.response):
            self.response_status = getattr(self.response, 'status_code', 0)

            try:
                self.json = self.response.json()
                if self.is_error:
                    raise DirectusException(self)
            except jsonlib.decoder.JSONDecodeError:
                self.json = {}

    async def gather_response(self):
        if inspect.iscoroutine(self.response):
            self.response = await self.response
            self.parse_response()

    def _parse_item_as_dict(self) -> dict:
        if isinstance(self.json['data'], list):
            return self.json['data'][0]
        return self.json['data']

    def _parse_item_as_object(self, collection: T) -> T:
        return collection(**self._parse_item_as_dict())

    def _parse_items_as_dict(self) -> List[dict]:
        if isinstance(self.json['data'], list):
            return self.json['data']
        return [self.json['data']]

    def _parse_items_as_objects(self, collection: T) -> List[T]:
        items_data = self._parse_items_as_dict()
        return TypeAdapter(List[collection]).validate_python(items_data)

    @property
    def item(self) -> Union[dict[Any, Any], Any, None]:  # noqa
        if "data" not in self.json or self.json['data'] in [None, [], {}]:
            return None
        if self.collection:
            return self._parse_item_as_object(self.collection)
        return self._parse_item_as_dict()

    def item_as(self, collection: T) -> Union[T, None]:  # noqa
        item_data = self._parse_item_as_dict()
        return None if item_data is None else collection(**item_data)

    def item_as_dict(self) -> Union[dict, None]:  # noqa
        if "data" not in self.json or self.json['data'] in [None, [], {}]:
            return None
        return self._parse_item_as_dict()

    @property
    def items(self) -> Union[List[dict[Any, Any]], Any, None]:  # noqa
        if "data" not in self.json or self.json['data'] in [None, [], {}]:
            return None
        if self.collection:
            return self._parse_items_as_objects(self.collection)
        return self._parse_items_as_dict()

    def items_as(self, collection: T) -> Union[List[T], None]:  # noqa
        items_data = self._parse_items_as_dict()
        return None if items_data is None else TypeAdapter(List[collection]).validate_python(items_data)

    def items_as_dict(self) -> Union[List[dict], None]:  # noqa
        if "data" not in self.json or self.json['data'] in [None, [], {}]:
            return None
        return self._parse_items_as_dict()

    @property
    def total_count(self) -> int:
        if "meta" in self.json and "total_count" in self.json['meta']:
            return self.json['meta']['total_count']

    @property
    def filtered_count(self) -> int:
        if "meta" in self.json and "filter_count" in self.json['meta']:
            return self.json['meta']['filter_count']

    @property
    def status_code(self) -> int:
        return self.response_status

    @property
    def is_success(self) -> bool:
        return self.status_code in [200, 201, 204, 304]

    @property
    def is_error(self) -> bool:
        return not self.is_success

    @property
    def errors(self) -> list:
        if self.is_error and "errors" in self.json:
            return self.json['errors']
        else:
            return []

    def to_json(self):
        data = {
            "response_status": self.response_status,
            "query": self.query,
            # "collection": self.collection,
            "json": self.json
        }

        if self.collection:
            print(".".join([self.collection.__module__, self.collection.__name__]))

        return jsonlib.dumps(data)

    @classmethod
    def from_json(cls, json_data: str):
        data = jsonlib.loads(json_data)

        new_obj = cls()

        new_obj.response_status = data['response_status']
        new_obj.query = data['query']
        # new_obj.collection = data['collection']
        new_obj.json = data['json']

        return new_obj

    def get_explanation(self, show_headers=True, show_cookies=True) -> dict:
        needed_data = {}

        ''' Query '''

        needed_data["Query"] = {
            k: (v.get_explanation() if hasattr(v, 'get_explanation') else v) for k, v in self.query.items()
        }

        ''' Request '''

        resp_request = self.response.request

        request_data = {
            "method": resp_request.method,
            "url": str(resp_request.url)
        }

        # headers
        if show_headers:
            request_data["headers"] = self.response.headers.multi_items()

        # extensions
        request_data["extensions"] = resp_request.extensions

        needed_data["Request"] = request_data

        ''' Response '''

        response_data = {
            "status_code": self.response.status_code
        }

        # headers
        if show_headers:
            response_data["headers"] = self.response.headers.multi_items()

        # cookies
        if show_cookies:
            cookies = self.response.cookies

            response_data["cookies"] = [
                f"<Cookie {cookie.name}={cookie.value} for {cookie.domain} />"
                for cookie in cookies.jar
            ]

        # data
        response_data.update(self.json)

        needed_data["Response"] = response_data

        return needed_data

    def print_explanation(self, show_headers=True, show_cookies=True):
        console = Console()
        needed_data = self.get_explanation(show_headers=show_headers, show_cookies=show_cookies)

        console.print("///////// --------- Start DirectusResponse explanation --------- /////////", style="bold")

        for k, v in needed_data.items():
            console.print(k, style="bold")

            if k == "Query":
                for q_k, q_v in v.items():
                    console.print(q_k, style="bold")
                    console.print(q_v)
            else:
                console.print(v)

        console.print("///////// --------- End DirectusResponse explanation --------- /////////", style="bold")


class DirectusException(Exception):
    def __init__(self, response: DirectusResponse):
        self.response = response
        self.status_code = response.status_code
        self.message = None
        self.code = None

        if len(response.errors) > 0 and (
                "message" in response.errors[0]
                and "extensions" in response.errors[0]
                and "code" in response.errors[0]['extensions']
        ):
            self.message = response.errors[0]['message']
            self.code = response.errors[0]['extensions']['code']

    def __str__(self):
        return f"{self.code}: {self.message}"



from typing import Union, Optional

FITTING_OPTIONS = ["cover", "contain", "inside", "outside"]
IMAGE_FORMAT_OPTIONS = ["auto", "jpg", "png", "webp", "tiff"]


class ImageFileTransform:

    def __init__(
        self, 
        fit: Optional[str] = None, 
        width: Optional[int] = None, height: Optional[int] = None, 
        quality: Optional[int] = None, 
        withoutEnlargement: Optional[bool] = None, 
        img_format: Optional[str] = None, 
        **kwargs
    ):
        parameters = {
            "fit": fit if fit in FITTING_OPTIONS else None,
            "width": width,
            "height": height,
            "quality": quality,
            "withoutEnlargement": withoutEnlargement,
            "format": img_format if img_format in IMAGE_FORMAT_OPTIONS else None
        }

        self.parameters = {k:v for k,v in parameters.items() if v is not None}

        if kwargs:
            transforms = []

            for k, v in kwargs.items():
                parameter_list = [k, *(v if isinstance(v, list) else [v])]
                transforms.append(parameter_list)

            self.parameters["transforms"] = transforms



from typing import Optional

from httpx import AsyncClient

from py_directus import Directus


cached_directus_instances = dict[str, Directus]()

directus_session: AsyncClient = AsyncClient()
directus_session.timeout = 5
directus_session.headers.update({'Cache-Control': 'no-store'})

directus_url: str = None
# Client with administrator access
directus_admin: Optional[Directus] = None
# Public directus
directus_public: Optional[Directus] = None


async def async_init(directus_base_url: str, directus_admin_token: str = None):
    global directus_admin
    global directus_public
    global directus_url

    directus_url = directus_base_url
    directus_public = await Directus(directus_url, connection=directus_session)

    if directus_admin_token:
        directus_admin = await Directus(directus_url, token=directus_admin_token, connection=directus_session)
